class APIException(Exception):
    pass


class NoHostsAvailableException(APIException):
    def __init__(self, service_name):
        message = 'No hosts available for service: {service_name}'.format(service_name=service_name)
        super().__init__(message)


class UnfulfilledParameterException(APIException):
    def __init__(self, endpoint_path, unfulfilled_params):
        message = (
            'The {endpoint_path} endpoint '
            'was called without required parameters: {unfulfilled_params}'.format(
                endpoint_path=endpoint_path,
                unfulfilled_params=unfulfilled_params,
            )
        )
        super().__init__(message)



import collections
import logging
import random
from urllib import parse

import requests
from requests import adapters
from requests.packages.urllib3.util import retry

from apiron.exceptions import NoHostsAvailableException

LOGGER = logging.getLogger(__name__)

DEFAULT_CONNECTION_TIMEOUT = 1
DEFAULT_READ_TIMEOUT = 3

DEFAULT_CONNECTION_RETRIES = 1
DEFAULT_READ_RETRIES = 1
DEFAULT_TOTAL_RETRIES = 1
DEFAULT_STATUS_CODES_TO_RETRY_ON = range(500, 600)

Timeout = collections.namedtuple('Timeout', ['connection_timeout', 'read_timeout'])

DEFAULT_TIMEOUT = Timeout(connection_timeout=DEFAULT_CONNECTION_TIMEOUT, read_timeout=DEFAULT_READ_TIMEOUT)
DEFAULT_RETRY = retry.Retry(
    total=DEFAULT_TOTAL_RETRIES,
    connect=DEFAULT_CONNECTION_RETRIES,
    read=DEFAULT_READ_RETRIES,
    status_forcelist=DEFAULT_STATUS_CODES_TO_RETRY_ON,
)


class ServiceCaller:
    """
    A class for calling :mod:`services <apiron.service.base>`
    """

    @staticmethod
    def build_url(host, path):
        """
        Builds a valid URL from a host and path which may or may not have slashes in the proper place.
        Does not conform to `IETF RFC 1808 <https://tools.ietf.org/html/rfc1808.html>`_ but instead joins the host and path as given.
        Does not append any additional slashes to the final URL; just joins the host and path properly.

        :param str host:
            An HTTP host like ``'https://awesome-api.com/v2'``
        :param str path:
            The path to an endpoint on the host like ``'/some-resource/'``
        :return:
            The properly-joined URL of host and path, e.g. ``'https://awesome-api.com/v2/some-resource/'``
        :rtype:
            str
        """
        host += '/' if not host.endswith('/') else ''
        path = path.lstrip('/')

        return parse.urljoin(host, path)

    @staticmethod
    def get_adapted_session(adapter):
        """
        Mounts an adapter capable of communication over HTTP or HTTPS to the supplied session.

        :param adapter:
            A :class:`requests.adapters.HTTPAdapter` instance
        :return:
            The adapted :class:`requests.Session` instance
        """
        session = requests.Session()
        session.mount('http://', adapter)
        session.mount('https://', adapter)
        return session

    @staticmethod
    def get_required_headers(service, endpoint):
        """
        :param Service service:
            The service being called
        :param Endpoint endpoint:
            The endpoint being called
        :return:
            Headers required by the ``service`` and the ``endpoint`` being called
        :rtype:
            dict
        """
        headers = {}
        headers.update(service.required_headers)
        headers.update(endpoint.required_headers)
        return headers

    @staticmethod
    def choose_host(service):
        hosts = service.get_hosts()
        if not hosts:
            raise NoHostsAvailableException(service.service_name)
        return random.choice(hosts)

    @classmethod
    def build_request_object(
        cls,
        session,
        service,
        endpoint,
        method=None,
        path_kwargs=None,
        params=None,
        data=None,
        headers=None,
        cookies=None,
        auth=None,
    ):
        host = cls.choose_host(service=service)

        path_kwargs = path_kwargs or {}
        path = endpoint.get_formatted_path(**path_kwargs)

        merged_params = endpoint.get_merged_params(params)

        headers = headers or {}
        headers.update(cls.get_required_headers(service, endpoint))

        request = requests.Request(
            method=method or endpoint.default_method,
            url=cls.build_url(host, path),
            params=merged_params,
            data=data,
            headers=headers,
            cookies=cookies,
            auth=auth,
        )

        return session.prepare_request(request)

    @classmethod
    def call(
        cls,
        service,
        endpoint,
        method=None,
        path_kwargs=None,
        session=None,
        params=None,
        data=None,
        headers=None,
        cookies=None,
        auth=None,
        encoding=None,
        retry_spec=DEFAULT_RETRY,
        timeout_spec=DEFAULT_TIMEOUT,
        logger=None,
    ):
        """
        :param Service service:
            The service that hosts the endpoint being called
        :param Endpoint endpoint:
            The endpoint being called
        :param str method:
            The HTTP method to use for the call
        :param dict path_kwargs:
            Arguments to be formatted into the ``endpoint`` argument's ``path`` attribute
            (default ``None``)
        :param requests.Session session:
            (optional)
            An existing session, useful for making many calls in a single session
            (default ``None``)
        :param dict params:
            (optional)
            ``GET`` parameters to send to the endpoint
            (default ``None``)
        :param dict data:
            (optional)
            ``POST`` data to send to the endpoint.
            A :class:`dict` will be form-encoded, while a :class:`str` will be sent raw
            (default ``None``)
        :param dict headers:
            HTTP Headers to send to the endpoint
            (default ``None``)
        :param dict cookies:
            Cookies to send to the endpoint
            (default ``None``)
        :param auth:
            An object suitable for the :class:`requests.Request` object's ``auth`` argument
        :param str encoding:
            The codec to use when decoding the response.
            Default behavior is to have ``requests`` guess the codec.
            (default ``None``)
        :param urllib3.util.retry.Retry retry_spec:
            (optional)
            An override of the retry behavior for this call.
            (default ``Retry(total=1, connect=1, read=1, status_forcelist=[500-level status codes])``)
        :param Timeout timeout_spec:
            (optional)
            An override of the timeout behavior for this call.
            (default ``Timeout(connection_timeout=1, read_timeout=3)``)
        :param logging.Logger logger:
            (optional)
            An existing logger for logging from the proper caller for better correlation
        :return:
            The result of ``endpoint``'s :func:`format_response`
        :rtype: The type returned by ``endpoint``'s :func:`format_response`
        :raises requests.RetryError:
            if retry threshold exceeded due to bad HTTP codes (default 500 range)
        :raises requests.ConnectionError:
            if retry threshold exceeded due to connection or request timeouts
        """
        logger = logger or LOGGER

        if hasattr(endpoint, 'stub_response'):
            logger.info(
                'Stub call for endpoint defined by {}'
                .format(getattr(endpoint, 'endpoint_params', {}))
            )
            return endpoint.stub_response

        managing_session = False

        if not session:
            session = cls.get_adapted_session(adapters.HTTPAdapter(max_retries=retry_spec))
            managing_session = True

        request = cls.build_request_object(
            session,
            service,
            endpoint,
            method=method or endpoint.default_method,
            path_kwargs=path_kwargs,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            auth=auth,
        )

        logger.info('{method} {url}'.format(
            method=method or endpoint.default_method,
            url=request.url,
        ))

        response = session.send(
            request,
            timeout=(timeout_spec.connection_timeout, timeout_spec.read_timeout),
            stream=getattr(endpoint, 'streaming', False),
        )

        logger.info('{status} {url}{history}'.format(
            status=response.status_code,
            url=response.url,
            history=' ({} redirect(s))'.format(len(response.history)) if response.history else '',
        ))

        if managing_session:
            session.close()

        response.raise_for_status()

        if encoding:
            response.encoding = encoding

        return endpoint.format_response(response)



import collections
import logging
import warnings

from apiron.exceptions import UnfulfilledParameterException


LOGGER = logging.getLogger(__name__)


class Endpoint:
    """
    A basic service endpoint that responds with the default ``Content-Type`` for that endpoint
    """

    def __init__(self, path='/', default_method='GET', default_params=None, required_params=None):
        """
        :param str path:
            The URL path for this endpoint, without the protocol or domain
        :param str default_method:
            (Default ``'GET'``)
            The default method to use when calling this endpoint.
        :param dict default_params:
            The default parameters to use when calling this endpoint.
            Useful when an endpoint always or most often needs a base set of parameters supplied.
        :param required_params:
            An iterable of required parameter names.
            Calling an endpoint without its required parameters raises an exception.
        """
        self.default_method = default_method

        if '?' in path:
            warnings.warn(
                'Endpoint path may contain query parameters. '
                'Use the default_params or required_params attributes in the initialization of this endpoint, '
                'or the params argument when calling the endpoint instead.'.format(path),
                stacklevel=3,
            )

        self.path = path
        self.default_params = default_params or {}
        self.required_params = required_params or set()

    def format_response(self, response):
        """
        Extracts the appropriate type of response data from a :class:`requests.Response` object

        :param requests.Response response:
            The original response from :mod:`requests`
        :return:
            The response's text content
        :rtype:
            str
        """
        return response.text

    @property
    def required_headers(self):
        """
        Generates the headers that must be sent to this endpoint based on its attributes

        :return:
            Header name, header value pairs
        :rtype:
            dict
        """
        return {}

    def get_formatted_path(self, **kwargs):
        """
        Format this endpoint's path with the supplied keyword arguments

        :return:
            The fully-formatted path
        :rtype:
            str
        """
        return self.path.format(**kwargs)

    def get_merged_params(self, supplied_params=None):
        """
        Merge this endpoint's default parameters with the supplied parameters

        :param dict supplied_params:
            A dictionary of query parameter, value pairs
        :return:
            A dictionary of this endpoint's default parameters, merged with the supplied parameters.
            Any default parameters which have a value supplied are overridden.
        :rtype:
            dict
        :raises apiron.exceptions.UnfulfilledParameterException:
            When a required parameter for this endpoint is not a default param and is not supplied by the caller
        """
        supplied_params = supplied_params or {}

        empty_params = {
            param: supplied_params[param]
            for param in supplied_params
            if supplied_params[param] in (None, '')
        }
        if empty_params:
            warnings.warn(
                'The {path} endpoint '
                'was called with empty parameters: {empty_params}'.format(
                    path=self.path,
                    empty_params=empty_params,
                ),
                RuntimeWarning,
                stacklevel=4,
            )

        unfulfilled_params = {
            param
            for param in self.required_params
            if param not in supplied_params
            and param not in self.default_params
        }

        if unfulfilled_params:
            raise UnfulfilledParameterException(
                self.path,
                unfulfilled_params,
            )

        merged_params = self.default_params.copy()
        merged_params.update(supplied_params)
        return merged_params

    def __str__(self):
        return self.path


class JsonEndpoint(Endpoint):
    """
    An endpoint that returns :mimetype:`application/json`
    """

    def __init__(self, *args, path='/', default_method='GET', default_params=None, required_params=None, preserve_order=False):
        super().__init__(path=path, default_method=default_method, default_params=default_params, required_params=required_params)
        self.preserve_order = preserve_order

    def format_response(self, response):
        """
        Extracts JSON data from the response

        :param requests.Response response:
            The original response from :mod:`requests`
        :return:
            The response's JSON content
        :rtype:
            :class:`dict` if ``preserve_order`` is ``False``
        :rtype:
            :class:`collections.OrderedDict` if ``preserve_order`` is ``True``
        """

        return response.json(object_pairs_hook=collections.OrderedDict if self.preserve_order else None)

    @property
    def required_headers(self):
        return {'Accept': 'application/json'}


class StreamingEndpoint(Endpoint):
    """
    An endpoint that streams data incrementally
    """

    streaming = True

    def format_response(self, response):
        """
        Stream response in chunks

        :param requests.Response response:
            The original response from :mod:`requests`
        :return:
            The response's content
        :rtype:
            generator
        """

        return response.iter_content(chunk_size=None)


class StubEndpoint:
    """
    A stub endpoint designed to return a pre-baked response

    The intent is to allow for a service to be implemented
    before the endpoint is complete.
    """

    def __init__(self, stub_response=None, **kwargs):
        """
        :param stub_response:
            A pre-baked response, like 'stub response' or {'stub': 'response'}
        :param **kwargs:
            Arbitrary parameters that can match the intended real endpoint.
            These don't do anything for the stub but streamline the interface.
        """
        self.endpoint_params = kwargs if kwargs else {}
        self.stub_response = stub_response or 'stub for {}'.format(self.endpoint_params)



class Service:
    """
    A base class for low-level services.

    A service has a domain off of which one or more endpoints stem.
    """

    def get_hosts(self):
        """
        The fully-qualified hostnames that correspond to this service.
        These are often determined by asking a load balancer or service discovery mechanism.

        :return:
            The hostname strings corresponding to this service
        :rtype:
            list
        """
        return [self.domain]

    @property
    def required_headers(self):
        """
        Headers that are required when making calls to this :class:`Service`

        :return:
            Pairs of required header name, header value when calling this service
        :rtype:
            dict
        """
        return {}

    def __str__(self):
        return self.domain

    def __repr__(self):
        return '{klass}(domain={domain})'.format(klass=self.__class__.__name__, domain=self.domain)



from apiron.service.base import Service


class DiscoverableService(Service):
    """
    A Service whose hosts are determined via a host resolver.
    A host resolver is any class with a :func:`resolve` method
    that takes a service name as its sole argument and returns a
    list of host names that correspond to that service.
    """

    @classmethod
    def get_hosts(cls):
        return cls.host_resolver_class.resolve(cls.service_name)

    def __str__(self):
        return self.service_name

    def __repr__(self):
        return '{klass}(service_name={service_name}, host_resolver={host_resolver})'.format(
            klass=self.__class__.__name__,
            service_name=self.service_name,
            host_resolver=self.host_resolver_class.__name__,
        )



import os
from setuptools import setup, find_packages

with open('README.md') as readme_file:
    README = readme_file.read()

with open('requirements.txt') as requirements_file:
    REQUIREMENTS = [line.rstrip() for line in requirements_file if line != '\n']

with open('apiron/VERSION') as version_file:
    VERSION = version_file.readlines()[0].strip()

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='apiron',
    version=VERSION,
    author='Ithaka Harbors, Inc.',
    author_email='grp_ithaka_apiron@ithaka.org',
    url='https://github.com/ithaka/apiron',
    license='MIT',

    description='apiron helps you cook a tasty client for RESTful APIs. Just don\'t wash it with SOAP.',
    long_description=README,
    long_description_content_type='text/markdown',

    packages=find_packages(),
    include_package_data=True,

    install_requires=REQUIREMENTS,

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'License :: OSI Approved :: MIT License',
    ],
)
